/// \file
/// \brief \b [Internal] Weighted graph.  I'm assuming the indices are complex map types, rather than sequential numbers (which could be implemented much more efficiently).
///
/// This file is part of RakNet Copyright 2003 Kevin Jenkins.
///
/// Usage of RakNet is subject to the appropriate license agreement.
/// Creative Commons Licensees are subject to the
/// license found at
/// http://creativecommons.org/licenses/by-nc/2.5/
/// Single application licensees are subject to the license found at
/// http://www.rakkarsoft.com/SingleApplicationLicense.html
/// Custom license users are subject to the terms therein.
/// GPL license users are subject to the GNU General Public
/// License as published by the Free
/// Software Foundation; either version 2 of the License, or (at your
/// option) any later version.

#ifndef __WEIGHTED_GRAPH_H
#define __WEIGHTED_GRAPH_H

#include "DS_OrderedList.h"
#include "DS_Map.h"
#include "DS_Heap.h"
#include "DS_Queue.h"
#include <assert.h>
#ifdef _DEBUG
#include <stdio.h>
#endif

/// The namespace DataStructures was only added to avoid compiler errors for commonly named data structures
/// As these data structures are stand-alone, you can use them outside of RakNet for your own projects if you wish.
namespace DataStructures
{
	template <class node_type, class weight_type, bool allowUnlinkedNodes>
	class RAK_DLL_EXPORT WeightedGraph
	{
	public:
		static void IMPLEMENT_DEFAULT_COMPARISON(void) {DataStructures::defaultMapKeyComparison<node_type>(node_type(),node_type());}

		WeightedGraph();
		~WeightedGraph();
		void AddNode(node_type node);
		void RemoveNode(node_type node);
		void AddConnection(node_type node1, node_type node2, weight_type weight);
		void RemoveConnection(node_type node1, node_type node2);
		void Print(void);
		void Clear(void);
		bool GetShortestPath(node_type *destArray, unsigned *destArrayLength, node_type startNode, node_type endNode, weight_type INFINITE_WEIGHT);
		// TODO - put protected back when done testing.
	//protected:
		void ClearDijkstra(void);
		void GenerateDisjktraMatrix(node_type startNode, weight_type INFINITE_WEIGHT);

		DataStructures::Map<node_type, DataStructures::Map<node_type, weight_type> *> adjacencyLists;


		// All these variables are for path finding with Dijkstra
		struct  
		{
			bool isValid;
			node_type rootNode;
			DataStructures::OrderedList<node_type, node_type> costMatrixIndices;
			weight_type *costMatrix;
			node_type *leastNodeArray;
			unsigned costMatrixColumnWidth;
		} dijkstra;		
	};

	template <class node_type, class weight_type, bool allowUnlinkedNodes>
		WeightedGraph<node_type, weight_type, allowUnlinkedNodes>::WeightedGraph()
	{
		dijkstra.isValid=false;
		dijkstra.costMatrix=0;
	}

	template <class node_type, class weight_type, bool allowUnlinkedNodes>
		WeightedGraph<node_type, weight_type, allowUnlinkedNodes>::~WeightedGraph()
	{
		Clear();
	}

	template <class node_type, class weight_type, bool allowUnlinkedNodes>
		void WeightedGraph<node_type, weight_type, allowUnlinkedNodes>::AddNode(node_type node)
	{
		adjacencyLists.SetNew(node, new DataStructures::Map<node_type, weight_type>);
	}

	template <class node_type, class weight_type, bool allowUnlinkedNodes>
		void WeightedGraph<node_type, weight_type, allowUnlinkedNodes>::RemoveNode(node_type node)
	{
		unsigned i;
		DataStructures::Queue<node_type> removeNodeQueue;

		removeNodeQueue.Push(node);
		while (removeNodeQueue.Size())
		{
			delete adjacencyLists.Pop(removeNodeQueue.Pop());

			// Remove this node from all of the other lists as well
			for (i=0; i < adjacencyLists.Size(); i++)
			{
				adjacencyLists[i]->Delete(node);

				if (allowUnlinkedNodes==false && adjacencyLists[i]->Size()==0)
					removeNodeQueue.Push(adjacencyLists.GetKeyAtIndex(i));
			}
		}

		ClearDijkstra();
	}

	template <class node_type, class weight_type, bool allowUnlinkedNodes>
		void WeightedGraph<node_type, weight_type, allowUnlinkedNodes>::AddConnection(node_type node1, node_type node2, weight_type weight)
	{
		if (adjacencyLists.Has(node1)==false)
			AddNode(node1);
		adjacencyLists.Get(node1)->Set(node2, weight);
		if (adjacencyLists.Has(node2)==false)
			AddNode(node2);
		adjacencyLists.Get(node2)->Set(node1, weight);
	}

	template <class node_type, class weight_type, bool allowUnlinkedNodes>
		void WeightedGraph<node_type, weight_type, allowUnlinkedNodes>::RemoveConnection(node_type node1, node_type node2)
	{
		adjacencyLists.Get(node2)->Delete(node1);
		adjacencyLists.Get(node1)->Delete(node2);

		if (allowUnlinkedNodes==false) // If we do not allow unlinked nodes, then if there are no connections, remove the node
		{
			if (adjacencyLists.Get(node1)->Size()==0)
				RemoveNode(node1); // Will also remove node1 from the adjacency list of node2
			if (adjacencyLists.Get(node2)->Size()==0)
				RemoveNode(node2); // Will also remove node1 from the adjacency list of node2
		}

		ClearDijkstra();
	}

	template <class node_type, class weight_type, bool allowUnlinkedNodes>
		void WeightedGraph<node_type, weight_type, allowUnlinkedNodes>::Clear(void)
	{
		unsigned i;
		for (i=0; i < adjacencyLists.Size(); i++)
			delete adjacencyLists[i];
		adjacencyLists.Clear();

		ClearDijkstra();
	}

	template <class node_type, class weight_type, bool allowUnlinkedNodes>
		bool WeightedGraph<node_type, weight_type, allowUnlinkedNodes>::GetShortestPath(node_type *destArray, unsigned *destArrayLength, node_type startNode, node_type endNode, weight_type INFINITE_WEIGHT)
	{
		if (startNode==endNode)
		{
			destArray[0]=startNode;
			*destArrayLength=1;
			return true;
		}

		if (dijkstra.isValid==false || dijkstra.rootNode!=startNode)
		{
			ClearDijkstra();
			GenerateDisjktraMatrix(startNode, INFINITE_WEIGHT);
		}
		
		// return the results
		bool objectExists;
		unsigned col,row;
		weight_type currentWeight;
		DataStructures::Queue<node_type> outputQueue;
		col=dijkstra.costMatrixIndices.GetIndexFromKey(endNode, &objectExists);
		if (dijkstra.costMatrixIndices.Size()<=2)
		{
			assert(0);
			return false;
		}
		if (objectExists==false)
		{
			assert(0);
			return false;
		}		
		node_type vertex;
		row=dijkstra.costMatrixIndices.Size()-2;
		currentWeight=dijkstra.costMatrix[row*adjacencyLists.Size() + col];
		vertex=endNode;
		outputQueue.PushAtHead(vertex);
		row--;
		while (1)
		{
			while (dijkstra.costMatrix[row*adjacencyLists.Size() + col] == currentWeight)
			{
				if (row==0)
				{
					if (destArrayLength)
						*destArrayLength=outputQueue.Size()+1;
					destArray[0]=startNode;
					for (col=0; outputQueue.Size(); col++)
						destArray[col+1]=outputQueue.Pop();
					return true;
				}
				--row;
			}

			vertex=dijkstra.leastNodeArray[row];
			outputQueue.PushAtHead(vertex);
			if (row==0)
				break;
			col=dijkstra.costMatrixIndices.GetIndexFromKey(vertex, &objectExists);
			currentWeight=dijkstra.costMatrix[row*adjacencyLists.Size() + col];
		}
		if (destArrayLength)
			*destArrayLength=outputQueue.Size()+1;
		destArray[0]=startNode;
		for (col=0; outputQueue.Size(); col++)
			destArray[col+1]=outputQueue.Pop();
		return true;
	}

	template <class node_type, class weight_type, bool allowUnlinkedNodes>
		void WeightedGraph<node_type, weight_type, allowUnlinkedNodes>::GenerateDisjktraMatrix(node_type startNode, weight_type INFINITE_WEIGHT)
	{
		dijkstra.costMatrix = new weight_type[adjacencyLists.Size() * adjacencyLists.Size()];
		dijkstra.leastNodeArray = new node_type[adjacencyLists.Size()];

		node_type currentNode;
		unsigned col, row, row2, openSetIndex;
		node_type adjacentKey;
		unsigned adjacentIndex;
		weight_type edgeWeight, currentNodeWeight, adjacentNodeWeight;
		DataStructures::Map<node_type, weight_type> *adjacencyList;
		DataStructures::Heap<weight_type, node_type, false> minHeap;
		DataStructures::Map<node_type, weight_type> openSet;

		for (col=0; col < adjacencyLists.Size(); col++)
		{
			// This should be already sorted, so it's a bit inefficient to do an insertion sort, but what the heck
			dijkstra.costMatrixIndices.InsertElement(adjacencyLists.GetKeyAtIndex(col),adjacencyLists.GetKeyAtIndex(col));
		}
		for (col=0; col < adjacencyLists.Size() * adjacencyLists.Size(); col++)
			dijkstra.costMatrix[col]=INFINITE_WEIGHT;
		currentNode=startNode;
		row=0;
		currentNodeWeight=0;
		dijkstra.rootNode=startNode;

		// Clear the starting node column
		adjacentIndex=adjacencyLists.GetIndexAtKey(startNode);
		for (row2=0; row2 < adjacencyLists.Size(); row2++)
			dijkstra.costMatrix[row2*adjacencyLists.Size() + adjacentIndex]=0;

		while (row < adjacencyLists.Size()-1)
		{
			adjacencyList = adjacencyLists.Get(currentNode);
			// Go through all connections from the current node.  If the new weight is less than the current weight, then update that weight.
			for (col=0; col < adjacencyList->Size(); col++)
			{
				edgeWeight=(*adjacencyList)[col];
				adjacentKey=adjacencyList->GetKeyAtIndex(col);
				adjacentIndex=adjacencyLists.GetIndexAtKey(adjacentKey);
				adjacentNodeWeight=dijkstra.costMatrix[row*adjacencyLists.Size() + adjacentIndex];

				if (currentNodeWeight + edgeWeight < adjacentNodeWeight)
				{
					// Update the weight for the adjacent node
					for (row2=row; row2 < adjacencyLists.Size(); row2++)
						dijkstra.costMatrix[row2*adjacencyLists.Size() + adjacentIndex]=currentNodeWeight + edgeWeight;
					openSet.Set(adjacentKey, currentNodeWeight + edgeWeight);
				}
			}

			// Find the lowest in the open set
			minHeap.Clear();
			for (openSetIndex=0; openSetIndex < openSet.Size(); openSetIndex++)
				minHeap.Push(openSet[openSetIndex], openSet.GetKeyAtIndex(openSetIndex));

			currentNodeWeight=minHeap.PeekWeight(0);
			dijkstra.leastNodeArray[row]=minHeap.Pop(0);
			currentNode=dijkstra.leastNodeArray[row];
			openSet.Delete(currentNode);
/*
			unsigned i,j;
			for (i=0; i < adjacencyLists.Size(); i++)
			{
				for (j=0; j < adjacencyLists.Size(); j++)
				{
					printf("%2i ", dijkstra.costMatrix[i*adjacencyLists.Size() + j]);
				}
				printf("Node=%i", dijkstra.leastNodeArray[i]);
				printf("\n");
			}
*/				
			row++;
		}
		unsigned i,j;
		for (i=0; i < adjacencyLists.Size()-1; i++)
		{
			for (j=0; j < adjacencyLists.Size(); j++)
			{
				printf("%2i ", dijkstra.costMatrix[i*adjacencyLists.Size() + j]);
			}
			printf("Node=%i", dijkstra.leastNodeArray[i]);
			printf("\n");
		}
		dijkstra.isValid=true;
	}

	template <class node_type, class weight_type, bool allowUnlinkedNodes>
		void WeightedGraph<node_type, weight_type, allowUnlinkedNodes>::ClearDijkstra(void)
	{
		if (dijkstra.isValid)
		{
			dijkstra.isValid=false;
			delete [] dijkstra.costMatrix;
			delete [] dijkstra.leastNodeArray;
			dijkstra.costMatrixIndices.Clear();
		}
	}

	template <class node_type, class weight_type, bool allowUnlinkedNodes>
		void WeightedGraph<node_type, weight_type, allowUnlinkedNodes>::Print(void)
	{
#ifdef _DEBUG
		unsigned i,j;
		for (i=0; i < adjacencyLists.Size(); i++)
		{
			printf("%i: ", i);

			if (adjacencyLists[i]->Size()==0)
				printf("<Empty>");
			else
			{
				for (j=0; j < adjacencyLists[i]->Size(); j++)
					printf("%i (%.2f) ", adjacencyLists.GetIndexAtKey(adjacencyLists[i]->GetKeyAtIndex(j)), (float) adjacencyLists[i]->operator[](j) );
			}

			printf("\n");
		}
#endif
	}
}

#endif
